"""Physical equations and models for building a simulation environment"""

import jax.numpy as jnp
import jax
from jax import grad, jit, vmap, lax
from functools import partial

from typing import NamedTuple, Tuple

from solarcarsim.noise import (
    fractal_noise_1d,
    generate_wind_field,
)


class MotorParams(NamedTuple):
    kv: float
    kt: float
    resistance: float
    friction_coeff: float
    iron_coeff: float


class BatteryParams(NamedTuple):
    shape: Tuple[int, int]  # (series,parallel) array of batteries
    resistance: float  # ohms
    initial_energy: float  # joules


class CarParams(NamedTuple):
    """Physical Data for Solar Car Parameters"""

    mass: float = 800  # kg
    frontal_area: float = 1.3  # m^2
    drag_coeff: float = 0.18  # drag coefficient, dimensionless
    rolling_coeff: float = 0.002  # rolling resistance.
    moter_eff: float = 0.93  # 0 < x < 1 scaling factor
    wheel_radius: float = 0.23  # wheel radius in meters
    max_speed: float = 30.0  # m/s top speed
    solar_area: float = 5.0  # m^2, typically 5.0
    solar_eff: float = 0.20  # 0 < x < 1, typically ~.25
    n_motors: int = 2  # how many motors we have.
    motor: MotorParams = MotorParams(
        8.43, 1.1, 100.0, 0.001, 0.001
    )  # mitsuba m2090 estimate
    battery: BatteryParams = BatteryParams(
        (36, 19), 0.0126, 66.6e3
    )  # freebasing 50s pack.


def DefaultCar() -> CarParams:
    """Creates a basic car"""
    return CarParams(1000, 1.3, 0.18, 0.002, 0.85, 5.0, 0.23)


# some physics equations using jax


@jit
def normal_force(mass, theta):
    return mass * 9.8 * jnp.cos(theta)


@jit
def downslope_force(mass, theta):
    return mass * 9.8 * jnp.sin(theta)


@partial(jit, static_argnames=["crr"])
def rolling_force(mass, theta, crr):
    return normal_force(mass, theta) * crr


@partial(jit, static_argnames=["area", "cd", "rho"])
def drag_force(u, area, cd, rho):
    return 0.5 * rho * jnp.pow(u, 2) * cd * area


# we can use those forces above to determine what forces we have to overcome. Sum(F)=0


# @partial(jit, static_argnums=(2,))
@jit
def bldc_power_draw(torque, velocity, params: MotorParams):
    """
    Approximates power draw of a BLDC motor outputting a torque at a given velocity

    Args:
        torq: Applied force in Newton/meters
        velocity: Angular velocity in rad/s
        resistance: Motor phase resistance (ohms)
        kt: Torque constant (N⋅m/A)
        friction_coeff: Mechanical friction coefficient
        iron_loss_coeff: Iron loss coefficient (core losses)

    Returns:
        Total electrical power draw in Watts
    """

    # Current required for torque (simplified relationship)
    current = torque / params.kt

    # Copper losses (I²R)
    copper_losses = params.resistance * current**2
    # Mechanical friction losses
    friction_losses = params.friction_coeff * velocity**2
    # Iron losses (simplified model - primarily dependent on speed)
    iron_losses = params.iron_coeff * velocity**2
    # Mechanical power output
    mechanical_power = torque * velocity

    # Total electrical power input
    total_power = mechanical_power + copper_losses + friction_losses + iron_losses

    return total_power


# @partial(jit, static_argnames=['resistance', 'kt', 'kv', 'vmax', 'Cf'])
@jit
def bldc_torque(velocity, current_limit, resistance, kt, kv, vmax, Cf):
    bemf = velocity / kv
    v_avail = jnp.clip(vmax - bemf, 0.0, vmax)
    current = jnp.clip(v_avail / resistance, 0.0, current_limit)

    torque = kt * current
    friction_torque = Cf * velocity
    net_torque = jnp.maximum(torque - friction_torque, 0.0)
    stall_torque = kt * current_limit
    return jnp.where(velocity < 0.01, stall_torque, net_torque)


@partial(
    jit,
    static_argnums=(
        1,
        2,
    ),
)
def battery_powerloss(current, cell_r, battery_shape: Tuple[int, int]):
    r_array = jnp.full(battery_shape, cell_r)
    branch_current = current / battery_shape[1]
    I_array = jnp.full(battery_shape, branch_current)
    cell_Ploss = jnp.square(I_array) * r_array
    return jnp.sum(cell_Ploss)


def forward(state, timestep, control, params: CarParams):
    # state is (position, time, energy)
    # control is velocity
    # timestep is >0 time to advance
    # params is the params dictionary.
    # returns the next state with (position', time + timestep, energy')
    # TODO: terrain, weather, solar

    # determine the forces acting on the car.
    dragf = drag_force(control, params.frontal_area, params.drag_coeff, 1.184)
    rollf = rolling_force(params.mass, 0, params.rolling_coeff)
    hillforce = downslope_force(params.mass, 0)
    totalf = dragf + rollf + hillforce
    # determine the power needed to make this force
    tau = params.wheel_radius * totalf
    pdraw = bldc_power_draw(tau, control, params.motor)
    net_power = 0 - pdraw  # watts aka j/s

    # TODO: calculate battery-based power losses.
    # TODO: support regenerative braking when going downhill
    # TODO: delta x = cos(theta) * velocity * timestep

    new_state = jnp.array(
        [
            state[0] + control * timestep,
            state[1] + timestep,
            state[2] + net_power * timestep,
        ]
    )
    return new_state


def make_environment(seed):
    """Generate a race environment: terrain function, wind function, wrapped forward function."""
    key, subkey = jax.random.split(seed)
    wind = generate_wind_field(subkey, 10000, 600, spatial_scale=1000)
    key, subkey = jax.random.split(key)
    slope = fractal_noise_1d(subkey, 10000, scale=1200, height_scale=0.08)
    elevation = jnp.cumsum(slope)
    # elevation = generate_elevation_profile(subkey, 10000, height_variation=40.0, scale=1200, octaves=5)
    # slope = jnp.arctan(jnp.diff(elevation, prepend=100.0)) # rise/run

    return wind, elevation, slope

